def serialize_suggestions_values_response(result):
    try:
        field_values = result["aggregations"]["values"]["buckets"]
    except KeyError:
        field_values = []
    return field_values


def serialize_mappings(mappings):
    result = []
    for field in mappings:
        result.append({"name": field.name, "type": field.elastic_field_type})
    return result



import json

from django.http import HttpResponse
from django.views.generic.base import View

from .config import get_dql_config
from .exceptions import IndexNotSpecified, SchemaError
from .schema import get_schema_instance
from .serializers import serialize_mappings


class BaseAPIView(View):
    def _get_index(self, request):
        dql_config = get_dql_config()
        accept_param = dql_config.accept_index_param
        index = request.GET.get("index") if accept_param else dql_config.default_index
        if not index:
            raise IndexNotSpecified("index not specified")
        return index

    def _error(self, message, status_code=400):
        return HttpResponse(
            content=json.dumps({"error": message}, indent=2),
            content_type='application/json; charset=utf-8',
            status=status_code
        )


class MappingsAPIView(BaseAPIView):
    http_method_names = ['get']

    def get(self, request, *args, **kwargs):
        try:
            index = self._get_index(request)
            schema_instance = get_schema_instance(index)
            mappings = schema_instance.get_mappings()
            serialized_mappings = serialize_mappings(mappings)
        except (SchemaError, IndexNotSpecified) as exception:
            return self._error(str(exception))
        return HttpResponse(
            content=json.dumps(serialized_mappings, indent=2),
            content_type='application/json; charset=utf-8',
        )


class SuggestionsAPIView(BaseAPIView):
    http_method_names = ['get']

    def get(self, request, field, *args, **kwargs):
        # TODO: add pagination
        try:
            index = self._get_index(request)
            schema_instance = get_schema_instance(index)
            search = request.GET.get("search")
            suggestions = schema_instance.suggestions(field, search=search)
        except (SchemaError, IndexNotSpecified) as exception:
            return self._error(message=str(exception))
        return HttpResponse(
            content=json.dumps(suggestions, indent=2),
            content_type='application/json; charset=utf-8',
        )



class AbstractQuery(object):
    invert = False

    def generate(self, field_name, value):
        raise NotImplementedError


class EqualsQuery(AbstractQuery):
    invert = False

    def generate(self, field_name, value):
        return {
                   "match": {
                       field_name: value
                   }
               }, self.invert


class NotEqualsQuery(AbstractQuery):
    invert = True

    def generate(self, field_name, value):
        return {
                   "match": {
                       field_name: value
                   }
               }, self.invert


class GtQuery(AbstractQuery):
    invert = False

    def generate(self, field_name, value):
        return {
                   "range": {
                       field_name: {
                           "gt": value
                       }
                   }
               }, self.invert


class GteQuery(AbstractQuery):
    invert = False

    def generate(self, field_name, value):
        return {
                   "range": {
                       field_name: {
                           "gte": value
                       }
                   }
               }, self.invert


class LtQuery(AbstractQuery):
    invert = False

    def generate(self, field_name, value):
        return {
                   "range": {
                       field_name: {
                           "lt": value
                       }
                   }
               }, self.invert


class LteQuery(AbstractQuery):
    invert = False

    def generate(self, field_name, value):
        return {
                   "range": {
                       field_name: {
                           "lte": value
                       }
                   }
               }, self.invert


class ContainsQuery(AbstractQuery):
    invert = False

    def generate(self, field_name, value):
        return {
                   "wildcard": {
                       field_name: {
                           "value": "*" + value + "*"
                       }
                   }
               }, self.invert


class NotContainsQuery(AbstractQuery):
    invert = True

    def generate(self, field_name, value):
        return {
                   "wildcard": {
                       field_name: {
                           "value": "*" + value + "*"
                       }
                   }
               }, self.invert


class InQuery(AbstractQuery):
    invert = False

    def generate(self, field_name, value):
        return [{"match_phrase": {field_name, phrase}} for phrase in value], self.invert


class NotInQuery(AbstractQuery):
    invert = True

    def generate(self, field_name, value):
        return [{"match_phrase": {field_name, phrase}} for phrase in value], self.invert


class StartsWithQuery(AbstractQuery):
    invert = False

    def generate(self, field_name, value):
        return {
                   "prefix": {
                       field_name: {
                           "value": value
                       }
                   }
               }, self.invert


class EndsWithQuery(AbstractQuery):
    invert = False

    def generate(self, field_name, value):
        return {
                   "wildcard": {
                       field_name: {
                           "value": "*" + value
                       }
                   }
               }, self.invert


class NotStartsWithQuery(AbstractQuery):
    invert = True

    def generate(self, field_name, value):
        return {
                   "prefix": {
                       field_name: {
                           "value": value
                       }
                   }
               }, self.invert


class NotEndsWithQuery(AbstractQuery):
    invert = True

    def generate(self, field_name, value):
        return {
                   "wildcard": {
                       field_name: {
                           "value": "*" + value
                       }
                   }
               }, self.invert


class QueryBuilderFactory(object):
    operator_class_mapping = {'=': EqualsQuery,
                              '>': GtQuery,
                              '>=': GteQuery,
                              '<': LtQuery,
                              '<=': LteQuery,
                              '~': ContainsQuery,
                              'in': InQuery,
                              'startswith': StartsWithQuery,
                              'endswith': EndsWithQuery,
                              '!=': NotEqualsQuery,
                              '!~': NotContainsQuery,
                              'not in': NotInQuery,
                              'not startswith': NotStartsWithQuery,
                              'not endswith': NotEndsWithQuery,
                              }

    def get_query_builder(self, operator):
        query_generator_cls = self.operator_class_mapping.get(operator)
        return query_generator_cls()



from django.utils.module_loading import import_string
from djangoql.ast import Name
from elasticsearch import Elasticsearch

from elastic_dql.serializers import serialize_suggestions_values_response
from .config import get_dql_config
from .exceptions import SchemaError
from .field import FieldMapper
from .utils import build_field_name_from_parts


class SchemaFactory(object):
    instance = None
    include_indices = ('*',)
    exclude_indices = ()
    index_field_limits = {}

    @classmethod
    def get_instance(cls):
        """
            Get one instance of schema foctory.(singleton pattern)
            it is class method because this class can be inherited
        :return:
        """
        if not cls.instance:
            cls.instance = cls()
        return cls.instance

    def __init__(self):
        if self.include_indices and self.exclude_indices:
            raise SchemaError(
                'Either include_indices or exclude_indices can be specified, but not both',
            )
        if not (self.include_indices or self.exclude_indices):
            raise SchemaError(
                'One of include_indices or exclude_indices must be specified',
            )
        self.per_index_instance = {}

    def get_schema_instance(self, schema_cls, index):
        if self.excluded(index):
            raise SchemaError(
                "%s can't be used with %s index because it's excluded from it" % (
                    index,
                    self.__class__,
                ),
            )
        try:
            schema_instance = self.per_index_instance.get(index, None)
            if not schema_instance or isinstance(schema_instance, schema_cls):
                schema_instance = self._create_schema_instance(schema_cls, index)
        except Exception as exception:
            raise SchemaError(str(exception))
        return schema_instance

    def excluded(self, index):
        if self.include_indices:
            excluded = ("*" not in self.include_indices) and (index not in self.include_indices)
        else:
            excluded = index in self.exclude_indices
        return excluded

    def _create_schema_instance(self, schema_cls, index):
        fields_limit = self.index_field_limits.get(index, [])
        client = self._create_elastic_connection()
        schema_instance = schema_cls(client, index, fields_limit=fields_limit)
        self.per_index_instance[index] = schema_instance
        return schema_instance

    def _create_elastic_connection(self):
        dql_config = get_dql_config()
        elastic_connection = Elasticsearch(**dql_config.elastic_connection_params)
        return elastic_connection


class AbstractElasticDjangoQlSchema(object):
    pass


class ElasticDjangoQlSchema(AbstractElasticDjangoQlSchema):

    def __init__(self, client, index, fields_limit=None):
        self.client = client
        self.index = index
        self.fields_limit = fields_limit if fields_limit else []
        self.valid_properties_dict = {}

    def get_mappings(self):
        try:
            result = self.client.indices.get_mapping(index=self.index)
            mappings = result[self.index]["mappings"]
        except Exception as e:
            raise SchemaError(str(e))
        valid_properties = self._cache_properties(mappings)
        return valid_properties

    def suggestions(self, field_name, search=None):
        field = self.resolve_name(field_name)
        if not field:
            raise SchemaError("%s is not a valid field_name" % field_name)
        if not field.can_suggest_values():
            raise SchemaError("field %s hasn't type keyword" % field_name)
        query, aggregations = self._get_suggestions_query(field.name, search)
        try:
            result = self.client.search(index=self.index, query=query, aggregations=aggregations)
        except Exception as exception:
            raise SchemaError(str(exception))
        final_result = serialize_suggestions_values_response(result)
        return final_result

    def resolve_name(self, field_name):
        if isinstance(field_name, Name):
            field_name = build_field_name_from_parts(field_name.parts)
        return self._resolve_name_str(field_name)

    def _resolve_name_str(self, field_name):
        if not self.valid_properties_dict:
            self.get_mappings()
        field = self.valid_properties_dict.get(field_name)
        if not field:
            raise SchemaError("invalid field_name: %s" % field_name)
        return field

    def _cache_properties(self, properties):
        field_mapper = FieldMapper()
        must_be_limited_field_name = set(self.fields_limit)
        all_fields = field_mapper.get_properties(properties)
        valid_properties = list(filter(lambda field: field.name not in must_be_limited_field_name, all_fields))
        self.valid_properties_dict = {field.name: field for field in valid_properties}
        return valid_properties

    def _get_suggestions_query(self, field_name, search):
        # TODO: seprate query building from schema
        if search:
            query = {
                "wildcard": {
                    field_name: {
                        "value": "*" + search + "*"
                    }
                }
            }
        else:
            query = {
                "exists": {
                    "field": field_name
                }
            }

        aggregations = {
            "values": {
                "terms": {"field": field_name}
            }}
        return query, aggregations

    def validate(self, ast):
        pass


class SchemaHandler:
    schema_factory = None
    schema_cls = None
    initiated = False

    @staticmethod
    def initiate():
        dql_config = get_dql_config()
        if not SchemaHandler.initiated:
            schema_factory_cls = import_string(dql_config.schema_factory_path)
            SchemaHandler.schema_factory = schema_factory_cls.get_instance()
            SchemaHandler.schema_cls = import_string(dql_config.default_schema_path)
            SchemaHandler.initiated = True
        return SchemaHandler.schema_factory, SchemaHandler.schema_cls


def get_schema_instance(index):
    schema_factory, schema_cls = SchemaHandler().initiate()
    schema_instance = schema_factory.get_schema_instance(schema_cls, index)
    return schema_instance



def dot_join(first: str, secound: str) -> str:
    if not (isinstance(first, str) and isinstance(secound, str)):
        raise TypeError("both first and second args should have the str type")
    return f"{first}.{secound}"


def build_field_name_from_parts(parts):
    return '.'.join(parts)


def get_fields(fields_dict, base_name):
    """
        this method is not used yet
        :param fields_dict:
        :param base_name:
        :return:
    """
    fields = []
    for field_name, data in fields_dict.items():
        fields.append({"name": dot_join(base_name, field_name), "type": data["type"]})
    return fields


def get_properties(mappings, base_name=None):
    """
        this method is not used yet
        :param fields_dict:
        :param base_name:
        :return:
    """
    result = []
    if not mappings:
        return result
    properties = mappings.get("properties")
    if properties:
        for property_name, property_data in properties.items():
            prefix = dot_join(base_name, property_name) if base_name else property_name
            props = get_properties(property_data, base_name=prefix)
            result.extend(props)
        return result
    field_type = mappings.get("type")
    if field_type:
        result.append({"name": base_name, "type": field_type})
        fields_property = mappings.get("fields")
        if fields_property:
            fields = get_fields(fields_property, base_name)
            result.extend(fields)
    return result



from django.conf import settings

from .exceptions import ConfigError


class ElasticDqlConfig:
    __instance = None
    schema_factory_path = None
    default_schema_path = None
    default_index = None
    accept_index_param = None
    elastic_connection_params = None

    def __init__(self, django_settings, default_settings):
        self.django_settings = django_settings
        self.defaults = default_settings
        self._configure()

    def _configure(self):
        settings = self.defaults if not self.django_settings else self.django_settings
        self.schema_factory_path = settings.get("schema_factory", self.defaults.get("schema_factory"))
        self.default_schema_path = settings.get("default_schema", self.defaults.get("default_schema"))
        self.default_index = settings.get("default_index", self.defaults.get("default_index"))
        self.accept_index_param = settings.get("accept_index_param", self.defaults.get("accept_index_param"))
        if not (self.accept_index_param or self.default_index):
            raise ConfigError("if accept_index_param is false default_index must be specified")
        self.elastic_connection_params = settings.get("connection", self.defaults.get("connection"))

    @staticmethod
    def get_config_instance(django_settings, default_settings):
        if not ElasticDqlConfig.__instance:
            ElasticDqlConfig.__instance = ElasticDqlConfig(django_settings, default_settings)
        return ElasticDqlConfig.__instance


DEFAULTS = {
    "schema_factory": "elastic_dql.schema.SchemaFactory",
    "default_schema": "elastic_dql.schema.ElasticDjangoQlSchema",
    "default_index": None,
    "accept_index_param": True,  # if false default_index should be specified
    "connection": {
        "hosts": ["http://localhost"],
    }
}


def get_dql_config():
    ELASTIC_DQL_SETTINGS_KEY = "ELASTIC_DQL"
    ELASTIC_DQL_SETTINGS = getattr(settings, ELASTIC_DQL_SETTINGS_KEY, {})
    dql_config = ElasticDqlConfig.get_config_instance(ELASTIC_DQL_SETTINGS, DEFAULTS)
    return dql_config



from collections import defaultdict

from djangoql.ast import Logical
from djangoql.parser import DjangoQLParser

from .schema import get_schema_instance


def finalize_query(query, inverted):
    base_query = {"query": {
        "bool": {
            "filter": [query]
        }
    }}
    return base_query


def build_query(expr, schema_instance):
    base_query = {
        "bool": {}
    }
    if isinstance(expr.operator, Logical):
        left, left_invert = build_query(expr.left, schema_instance)
        right, right_invert = build_query(expr.right, schema_instance)
        if expr.operator.operator == 'or':
            base_query["bool"]["minimum_should_match"] = 1
            base_query["bool"]["should"] = [left, right]
        else:
            base_query["bool"]["filter"] = [left, right]
        return base_query, False

    field = schema_instance.resolve_name(expr.left)

    # if not field:
    #     That must be a reference to a model without specifying a field.
    #     Let's construct an abstract lookup field for it
    #     field = ElasticDjangoQlField(
    #       name=expr.left.parts[-1],
    #       nullable=True,
    #     )
    query, inverted = field.get_lookup(
        path=expr.left.parts[:-1],
        operator=expr.operator.operator,
        value=expr.right.value,
    )
    if inverted:
        base_query["bool"]["must_not"] = [query]
    else:
        base_query["bool"]["minimum_should_match"] = 1
        base_query["bool"]["should"] = [query]
    return base_query, inverted


def get_query(index, search):
    ast = DjangoQLParser().parse(search)
    schema_instance = get_schema_instance(index)
    schema_instance.validate(ast)
    query, inverted = build_query(ast, schema_instance)
    query = finalize_query(query, inverted)
    return query



from djangoql.exceptions import DjangoQLError


class ElasticDjangoQLError(DjangoQLError):
    pass


class SchemaError(ElasticDjangoQLError):
    pass


class FieldError(ElasticDjangoQLError):
    pass


class ConfigError(ElasticDjangoQLError):
    pass


class IndexNotSpecified(ElasticDjangoQLError):
    pass



from django.urls import path

from .views import MappingsAPIView, SuggestionsAPIView

urlpatterns = [
    path('mappings', MappingsAPIView.as_view(), name='mappings_api'),
    path('suggestions/<str:field>', SuggestionsAPIView.as_view(), name='suggestions_api'),
]


def get_urls():
    return urlpatterns



from datetime import datetime

from djangoql.compat import text_type

from .exceptions import FieldError
from .query_generator import QueryBuilderFactory
from .utils import dot_join


class ElasticDjangoQlField(object):
    name = None
    nullable = False
    field_type = None
    elastic_field_type = ""
    value_types = []
    valid_operators = (
        '=', '>', '>=', '<', '<=', '~', 'in', 'startswith', 'endswith', '!=', '!~', 'not in', 'not startswith',
        'not endswith')
    value_types_description = ''

    def __init__(self, name, nullable=False, field_type=None, elastic_field_type=None):
        self.name = name
        self.nullable = nullable
        self.elastic_field_type = elastic_field_type
        if field_type is not None:
            self.field_type = field_type

    def get_lookup_name(self):
        return self.name

    def get_lookup(self, path, operator, value):
        # field_name = '.'.join(path + [self.get_lookup_name()])
        field_name = self.get_lookup_name()
        if operator not in self.valid_operators:
            raise FieldError("operator %s is not valid for this type" % operator)
        formatted_value = self.format_value(value)
        query_builder_factory = QueryBuilderFactory()
        query_builder = query_builder_factory.get_query_builder(operator)
        query, invert = query_builder.generate(field_name, formatted_value)
        return query, invert

    def format_value(self, value):
        """
        override this method to clean value or change type, ...

        :param value:
        :return: formatted_value:
            formatted_value:
        """
        return self.field_type(value)

    def validate(self, value):
        if not self.nullable and value is None:
            raise FieldError(
                'Field %s is not nullable, '
                "can't compare it to None" % self.name,
            )
        if value is not None and type(value) not in self.value_types:
            if self.nullable:
                msg = (
                    'Field "{field}" has "nullable {field_type}" type. '
                    'It can be compared to {possible_values} or None, '
                    'but not to {value}'
                )
            else:
                msg = (
                    'Field "{field}" has "{field_type}" type. It can '
                    'be compared to {possible_values}, '
                    'but not to {value}'
                )
            raise FieldError(msg.format(
                field=self.name,
                field_type=self.field_type,
                possible_values=self.value_types_description,
                value=repr(value),
            ))

    def can_suggest_values(self):
        return False


class LongField(ElasticDjangoQlField):
    value_types = [int, float]
    field_type = int
    value_types_description = "numeric fields"


class FloatField(ElasticDjangoQlField):
    value_types = [int, float]
    field_type = float
    value_types_description = "numeric fields"


class TextType(ElasticDjangoQlField):
    value_types = [text_type]
    field_type = text_type
    value_types_description = "strings"


class DateField(ElasticDjangoQlField):
    value_types = [text_type]
    field_type = text_type
    value_types_description = 'dates in "YYYY-MM-DD" format'

    def validate(self, value):
        super(DateField, self).validate(value)
        try:
            datetime.strptime(value, '%Y-%m-%d').date()
        except:
            raise FieldError()
    # TODO: time format checking
    # pass


class BoolField(ElasticDjangoQlField):
    value_types = [bool]
    field_type = bool
    value_types_description = "booleans"


class KeywordField(ElasticDjangoQlField):
    value_types = [text_type]
    field_type = text_type

    def __init__(self, name, parent, *args, **kwargs):
        super(KeywordField, self).__init__(name, *args, **kwargs)
        self.parent = parent

    def can_suggest_values(self):
        return True


class FieldMapper(object):
    type_mapping = {
        "date": DateField,
        "keyword": KeywordField,
        "boolean": BoolField,
        "long": LongField,
        "text": TextType,
        "float": FloatField,
        "unsigned_long": LongField
    }

    def get_fields(self, fields_dict: dict, parent: ElasticDjangoQlField):
        fields = []
        for field_name, data in fields_dict.items():
            field_cls = self._get_field_cls(data["type"])
            field_name = dot_join(parent.name, field_name)
            field = field_cls(field_name, parent=parent, elastic_field_type=data["type"])
            fields.append(field)
        return fields

    def get_properties(self, mappings, base_name=None):
        result = []
        if not mappings:
            return result
        properties = mappings.get("properties")
        if properties:
            for property_name, property_data in properties.items():
                prefix = dot_join(base_name, property_name) if base_name else property_name
                props = self.get_properties(property_data, base_name=prefix)
                result.extend(props)
            return result
        field_type = mappings.get("type")
        if field_type:
            field_cls = self._get_field_cls(field_type)
            field = field_cls(base_name, elastic_field_type=field_type)
            result.append(field)
            fields_property = mappings.get("fields")
            if fields_property:
                fields = self.get_fields(fields_property, field)
                result.extend(fields)
        return result

    def _get_field_cls(self, property_type):
        field_cls = self.type_mapping.get(property_type)
        if not field_cls:
            raise FieldError("invalid field type: %s" % property_type)
        return field_cls



#!/usr/bin/env python

from setuptools import setup
from pathlib import Path
import elastic_dql

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

packages = ['elastic_dql']
requires = ["djangoql==0.17.1", "elasticsearch==8.1.0"]

setup(
    name='elastic-dql',
    packages=packages,
    install_requires=requires,
    version=elastic_dql.__version__,
    description='Elastic query language library - convering readable queries to elasticsearch query',
    long_description_content_type='text/markdown',
    long_description=long_description,
    author='Mohammad Ahadi',
    author_email='mohammadahadi27s@gmail.com',
    url='https://github.com/mohammad-ahadi/elastic-dql',
    license=open('LICENSE').readline().strip(),
    zip_safe=False,

)
